﻿import util.Observer;

/**
 * 一個Java樣式的Observable類別被用來代表觀察者 
 * 設計模式的「對象」。所有觀察者都必須實作Observer
 * 介面，並且透過addObserver( )來註冊並觀察該對象。
 */
class util.Observable {
  // 指出這個物件是否改變的旗標。
  private var changed:Boolean = false;
  // 記錄觀察者名單的陣列。
  private var observers:Array;

  /**
   * 建構子函數。
   */
  public function Observable (  ) {
    observers = new Array(  );
  }

  /**
   * 把一個觀察者加入Observers陣列。
   * @param   o   要被加入的觀察者。 
   */
  public function addObserver(o:Observer):Boolean {
    // 不能加入一個空的（null）觀察者。
    if (o == null) {
      return false;
    }

    // 不能再次加入相同的觀察者。
    for (var i:Number = 0; i < observers.length; i++) {
      if (observers[i] == o) {
        // 指定的觀察者已經在觀察名單中了，所以退出。 
        return false;
      }
    }

    // 把此觀察者加入觀察名單。
    observers.push(o);
    return true;
  }

  /**
   * 從觀察名單中移除一個觀察者。
   *
   * @param   o   要被移除的觀察者。
   */
  public function removeObserver(o:Observer):Boolean {
    // 找出並移出這個觀察者。
    var len:Number = observers.length;
    for (var i:Number = 0; i < len; i++) {
      if (observers[i] == o) {
        observers.splice(i, 1);
        return true;
      }
    }
    return false;
  }

  /**
   * 告知所有觀察者，「對象」已經改變了。
   *
   * @param   infoObj   包含要傳給觀察者的資料的物件
   */
  public function notifyObservers(infoObj:Object):Void {
    // 如果沒有資料要傳遞，就將infoObj設定成null。
    if (infoObj == undefined) {
      infoObj = null;
    }

    // 假若對象沒有改變，就不必通知觀察者。
    if (!changed) {
      return;
    }

    // 複製一份觀察者陣列。這是為了確保當我們 
    // 在處理此陣列時，不會更動到它的內容。
    var observersSnapshot:Array = observers.slice(0);

    // 已經處理好變化了，所以要清除changed旗標。
    clearChanged(  );

    // 對所有觀察者呼叫update( )。底下的迴圈採用倒數方式， 
    // 因為這樣比較快，而且正數或倒數對這個範例沒有差別。
    for (var i:Number = observersSnapshot.length-1; i >= 0; i--) {
      observersSnapshot[i].update(this, infoObj);
    }
  }

  /**
   * 移除觀察者名單裡面的所有觀察者。
   */
  public function clearObservers(  ):Void {
    observers = new Array(  );
  }

  /**
   *  指出對象產生了變化。
   */
  private function setChanged(  ):Void {
    changed = true;
  }

  /**
   * 指出對象沒有產生變化，或者已經向 
   * 它的觀察者傳達最新的改變。
   */
  private function clearChanged(  ):Void {
    changed = false;
  }

  /**
   * 檢查對象是否改變了。
   *
   * @return   如果有變化，則傳回true，
   *           這個值是由setChanged( )決定的。
   */
  public function hasChanged(  ):Boolean {
    return changed;
  }

  /**
   * 傳回觀察者名單中的觀察者總數。
   *
   * @return   整數：這個對象的觀察者總數。
   */
  public function countObservers(  ):Number {
    return observers.length;
  }
}